/*
 * tccext.c - Parse information received from tcc's external interface
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 Network Robots.
 * Distributed under the LGPL.
 */


#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>

#include "tccext.h"


#define MAX_LINE	65536	/* maximum input line */
#define MAX_BUF		1024	/* maximum token length */


#define _S(x) #x
#define S(x) _S(x)


static void *alloc_zero(size_t size)
{
    void *p;

    p = malloc(size);
    if (!p) {
	perror("malloc");
	exit(1);
    }
    memset(p,0,size);
    return p;
}


static TCCEXT_OFFSET *find_group(TCCEXT_CONTEXT *ctx,int number)
{
    TCCEXT_OFFSET *group;

    if (!number) return NULL;
    for (group = ctx->offset_groups; group; group = group->next)
	if (group->group_number == number) return group;
    fprintf(stderr,"offset group %d not found\n",number);
    exit(1);
}


static TCCEXT_FIELD parse_field(TCCEXT_CONTEXT *ctx,const char *line,int *len)
{
    TCCEXT_FIELD field;
    int base,pos;

    if (sscanf(line,"%i:%i:%i%n",&base,&field.offset,&field.length,&pos) < 3) {
	fprintf(stderr,"invalid field expression \"%s\"\n",line);
	exit(1);
    }
    field.offset_group = find_group(ctx,base);
    if (len) *len = pos;
    return field;
}


static void parse_offset(TCCEXT_CONTEXT *ctx,const char *line)
{
    TCCEXT_OFFSET *offset;
    int base,pos;

    offset = alloc_zero(TE_SIZEOF(ctx,offset,TCCEXT_OFFSET));
    if (sscanf(line,"%i = %i+(%n",&offset->group_number,&base,&pos) >= 2) {
	int pos2;

	offset->base = find_group(ctx,base);
	offset->field = parse_field(ctx,line+pos,&pos2);
	if (sscanf(line+pos+pos2+1,"<< %i)",&offset->shift_left) == 1) {
	    offset->next = ctx->offset_groups;
	    ctx->offset_groups = offset;
	    return;
	}
    }
    fprintf(stderr,"invalid offset expression \"%s\"\n",line);
    exit(1);
}


static TCCEXT_ACTIONS *find_actions(TCCEXT_CONTEXT *ctx,int index)
{
    TCCEXT_ACTIONS *actions;

    for (actions = ctx->actions; actions; actions = actions->next)
	if (actions->index == index) break;
    return actions;
}


static int nibble(char c)
{
    return c >= 'a' ? c-'a'+10 : c >= 'A' ? c-'A'+10 : c-'0';
}


static TCCEXT_MATCH *parse_match(TCCEXT_CONTEXT *ctx,const char *line,int *len)
{
    TCCEXT_MATCH *match;
    int pos;

    match = alloc_zero(TE_SIZEOF(ctx,match,TCCEXT_MATCH));
    match->field = parse_field(ctx,line,&pos);
    if (!strncmp(line+pos,"=0x",3)) {
	int nibbles = 0,bytes;
	const char *walk;
	uint8_t *p;

	while (line[pos+3] == '0') pos++; /* avoid overflow on 0x000000... */
	for (walk = line+pos+3; *walk && isxdigit(*walk); walk++)
	    nibbles++;
	bytes = (match->field.length+7) >> 3;
	match->data = alloc_zero(bytes);
	p = match->data+bytes-((nibbles+1) >> 1);
	walk = line+pos+3;
	if (nibbles & 1) {
	    *p++ = nibble(*walk++);
	    nibbles--;
	}
	while (nibbles) {
	    *p = nibble(*walk++) << 4;
	    *p++ |= nibble(*walk++);
	    nibbles -= 2;
	}
	match->next = NULL;
	match->context = ctx;
	if (len) *len = walk-line;
	return match;
    }
    fprintf(stderr,"invalid match \"%s\"\n",line);
    exit(1);
}


static void parse_rule(TCCEXT_CONTEXT *ctx,const char *line)
{
    TCCEXT_RULE *rule,**walk;
    TCCEXT_MATCH **next;
    int actions,pos;

    rule = alloc_zero(TE_SIZEOF(ctx,rule,TCCEXT_RULE));
    rule->matches = NULL;
    next = &rule->matches;
    while (*line && isdigit(*line)) {
	*next = parse_match(ctx,line,&pos);
	next = &(*next)->next;
	line += pos+1;
    }
    if (sscanf(line," action %i",&actions) != 1) {
	fprintf(stderr,"invalid actions \"%s\"\n",line);
	exit(1);
    }
    rule->actions = find_actions(ctx,actions);
    if (!rule->actions) {
	fprintf(stderr,"no actions %d\n",actions);
	exit(1);
    }
    rule->next = NULL;
    for (walk = &ctx->rules; *walk; walk =&(*walk)->next);
    *walk = rule;
}


static TCCEXT_QDISC *find_qdisc(TCCEXT_CONTEXT *ctx,int index)
{
    TCCEXT_BLOCK *block;
    TCCEXT_QDISC *qdisc;

    for (block = ctx->blocks; block; block = block->next)
	for (qdisc = block->qdiscs; qdisc; qdisc = qdisc->next)
	    if (qdisc->index == index) return qdisc;
    return NULL;
}


/* @@@ REMOVE THIS ! */
static TCCEXT_QDISC *fudge_qdisc(TCCEXT_CONTEXT *ctx,int index)
{
    TCCEXT_QDISC *qdisc;

    qdisc = alloc_zero(TE_SIZEOF(ctx,qdisc,TCCEXT_QDISC));
    qdisc->index = index;
    qdisc->classes = NULL;
    return qdisc;
}


static TCCEXT_PARAMETER *parse_parameters(TCCEXT_CONTEXT *ctx,const char **line)
{
    TCCEXT_PARAMETER *result = NULL,**next = &result;
    char *name;
    unsigned long value;
    int pos;

    while (sscanf(*line,"%as %lu%n",&name,&value,&pos) >= 2) {
	TCCEXT_PARAMETER *prm;

	if (!strcmp(name,"pragma")) {
	    free(name);
	    return result;
	}
	prm = alloc_zero(TE_SIZEOF(ctx,parameter,TCCEXT_PARAMETER));
	prm->name = name;
	prm->value = value;
	prm->next = NULL;
	*next = prm;
	next = &prm->next;
	*line += pos;
    }
    return result;
}


static TCCEXT_PRAGMA *parse_pragmas(const TCCEXT_CONTEXT *ctx,const char *line)
{
    TCCEXT_PRAGMA *pragmas,**last;
    char *s;
    int pos = 0,len;

    last = &pragmas;
    while (sscanf(line+pos,"%as%n",&s,&len) >= 1) {
	*last = alloc_zero(TE_SIZEOF(ctx,pragma,TCCEXT_PRAGMA));
	(*last)->pragma = s;
	last = &(*last)->next;
	pos += len;
    }
    *last = NULL;
    return pragmas;
}


static void parse_pragma(TCCEXT_CONTEXT *ctx,const char *line)
{
    ctx->pragmas = parse_pragmas(ctx,line);
}


static void parse_qdisc(TCCEXT_CONTEXT *ctx,const char *line)
{
    TCCEXT_QDISC *qdisc;
    int pos;

    qdisc = alloc_zero(TE_SIZEOF(ctx,qdisc,TCCEXT_QDISC));
    if (sscanf(line,"%i = %as%n",&qdisc->index,&qdisc->type,&pos) < 2) {
	fprintf(stderr,"unrecognized qdisc definition \"%s\"\n",line);
	exit(1);
    }
    if (find_qdisc(ctx,qdisc->index)) {
	fprintf(stderr,"duplicate qdisc index %d\n",qdisc->index);
	exit(1);
    }
    line += pos+1;
    qdisc->parameters = parse_parameters(ctx,&line);
    while (*line == ' ') line++;
    if (!strncmp(line,"pragma",6)) qdisc->pragmas = parse_pragmas(ctx,line);
    qdisc->next = ctx->blocks->qdiscs;
    ctx->blocks->qdiscs = qdisc;
}


static void parse_class(TCCEXT_CONTEXT *ctx,const char *line)
{
    TCCEXT_CLASS *class;
    int pos;

    class = alloc_zero(TE_SIZEOF(ctx,class,TCCEXT_CLASS));
    if (!sscanf(line,"%i =%n",&class->index,&pos)) {
	fprintf(stderr,"unrecognized class definition \"%s\"\n",line);
	exit(1);
    }
    class->parent_qdisc = ctx->blocks->qdiscs;
    line += pos+1;
    class->parameters = parse_parameters(ctx,&line);
    while (*line == ' ') line++;
    if (!strncmp(line,"pragma",6)) class->pragmas = parse_pragmas(ctx,line);
    class->next = class->parent_qdisc->classes;
    class->parent_qdisc->classes = class;
}


static void parse_block(TCCEXT_CONTEXT *ctx,const char *line)
{
    TCCEXT_BLOCK *block;
    char role[11];
    int pos;

    block = alloc_zero(TE_SIZEOF(ctx,block,TCCEXT_BLOCK));
    if (sscanf(line,"%as %10s%n",&block->name,role,&pos) < 2) {
	fprintf(stderr,"invalid block \"%s\"\n",line);
	exit(1);
    }
    if (!strcmp(role,"ingress")) block->role = tbr_ingress;
    else if (!strcmp(role,"egress")) block->role = tbr_egress;
    else {
	fprintf(stderr,"unrecognized role \"%s\"\n",role);
	exit(1);
    }
    block->qdiscs = NULL;
    block->pragmas = parse_pragmas(ctx,line+pos);
    block->next = ctx->blocks;
    ctx->blocks = block;
}


static TCCEXT_BUCKET *find_bucket(TCCEXT_CONTEXT *ctx,int index)
{
    TCCEXT_BUCKET *bucket;

    for (bucket = ctx->buckets; bucket; bucket = bucket->next)
	if (bucket->index == index) return bucket;
    fprintf(stderr,"no bucket %d\n",index);
    exit(1);
}


static void parse_bucket(TCCEXT_CONTEXT *ctx,const char *line)
{
    TCCEXT_BUCKET *bucket;
    unsigned long rate,depth,initial;
    int overflow,pos;

    /* @@@ %lu should allow hex too %i */
    bucket = alloc_zero(TE_SIZEOF(ctx,bucket,TCCEXT_BUCKET));
    if (sscanf(line,"%i = %lu %i %lu %lu %i%n",&bucket->index,&rate,
      &bucket->mpu,&depth,&initial,&overflow,&pos) < 6) {
	fprintf(stderr,"unrecognized bucket \"%s\"\n",line);
	exit(1);
    }
    bucket->rate = rate;
    bucket->depth = depth;
    bucket->initial_tokens = initial;
    bucket->overflow = overflow ? find_bucket(ctx,overflow) : NULL;
    bucket->pragmas = parse_pragmas(ctx,line+pos);
    bucket->next = ctx->buckets;
    ctx->buckets = bucket;
}


static TCCEXT_CLASS *parse_classes(TCCEXT_CONTEXT *ctx,const char **line)
{
    TCCEXT_CLASS *classes,**next = &classes;

    do {
	int qdisc,pos;

	(*line)++;
/* @@@ this should be just a lookup */
	*next = alloc_zero(TE_SIZEOF(ctx,class,TCCEXT_CLASS));
	if (sscanf(*line,"%i:%i%n",&qdisc,&(*next)->index,&pos) < 2 ) {
	    fprintf(stderr,"not a class \"%s\"\n",*line);
	    exit(1);
	}
	(*next)->parent_qdisc = find_qdisc(ctx,qdisc);
	if (!(*next)->parent_qdisc)
	    (*next)->parent_qdisc = fudge_qdisc(ctx,qdisc);
		/* need to do this if output contains only classifier */
	next = &(*next)->next;
	*line += pos;
    }
    while (**line == ',');
    return classes;
}


static TCCEXT_ACTION *parse_action(TCCEXT_CONTEXT *ctx,const char **line)
{
    TCCEXT_ACTION *action;
    int next,bucket,pos;

    while (**line == ' ') (*line)++;
    if (sscanf(*line,"action %i%n",&next,&pos) >= 1) {
	*line += pos+1;
	return find_actions(ctx,next)->action;
    }
    action = alloc_zero(TE_SIZEOF(ctx,action,TCCEXT_ACTION));
    if (!strncmp(*line,"class",4)) {
	action->type = tat_class;
	*line += 4;
	action->u.classes = parse_classes(ctx,line);
	return action;
    }
    if (!strncmp(*line,"drop",4)) {
	action->type = tat_drop;
	*line += 4;
	return action;
    }
    if (!strncmp(*line,"unspec",7)) {
	action->type = tat_unspec;
	*line += 7;
	return action;
    }
    if (sscanf(*line,"conform %i%n",&bucket,&pos) >= 1) {
	action->type = tat_conform;
	action->u.conform.bucket = find_bucket(ctx,bucket);
	*line += pos+1;
	action->u.conform.yes_action = parse_action(ctx,line);
	action->u.conform.no_action = parse_action(ctx,line);
	return action;
    }
    if (sscanf(*line,"count %i%n",&bucket,&pos) >= 1) {
	action->type = tat_count;
	action->u.count.bucket = find_bucket(ctx,bucket);
	*line += pos+1;
	action->u.count.action = parse_action(ctx,line);
	return action;
    }
    fprintf(stderr,"unrecognized action \"%s\"\n",*line);
    exit(1);
}


static void parse_actions(TCCEXT_CONTEXT *ctx,const char *line)
{
    TCCEXT_ACTIONS *actions;
    int pos;

    actions = alloc_zero(TE_SIZEOF(ctx,actions,TCCEXT_ACTIONS));
    if (!sscanf(line,"%i =%n",&actions->index,&pos)) {
	fprintf(stderr,"unrecognized action definition \"%s\"\n",line);
	exit(1);
    }
    line += pos+1;
    actions->action = parse_action(ctx,&line);
    actions->next = ctx->actions;
    ctx->actions = actions;
}


TCCEXT_CONTEXT *tccext_parse(FILE *file,TCCEXT_SIZES *sizes)
{
    TCCEXT_CONTEXT *ctx;
    char line[MAX_LINE];
    
    ctx = alloc_zero(__TE_SIZEOF(sizes ? sizes->context : 0,
      sizeof(TCCEXT_CONTEXT)));
    ctx->offset_groups = NULL;
    ctx->rules = NULL;
    ctx->blocks = NULL;
    if (sizes) ctx->sizes = *sizes;
    else memset(&ctx->sizes,0,sizeof(TCCEXT_SIZES));
    while (fgets(line,sizeof(line),file)) {
	char buf[MAX_BUF];
	char *here;
	int pos;

	here = strchr(line,'\n');
	if (here) *here = 0;
	here = strchr(line,'#');
	if (here) *here = 0;
	if (!*line) continue;
	if (!sscanf(line,"%" S(MAX_BUF) "s%n",buf,&pos)) {
	    fprintf(stderr,"unrecognized line \"%s\"\n",line);
	    exit(1);
	}
	if (!strcmp(buf,"pragma")) parse_pragma(ctx,line+pos+1);
	else if (!strcmp(buf,"block")) parse_block(ctx,line+pos+1);
	else if (!strcmp(buf,"qdisc")) parse_qdisc(ctx,line+pos+1);
	else if (!strcmp(buf,"class")) parse_class(ctx,line+pos+1);
	else if (!strcmp(buf,"offset")) parse_offset(ctx,line+pos+1);
	else if (!strcmp(buf,"bucket")) parse_bucket(ctx,line+pos+1);
	else if (!strcmp(buf,"action")) parse_actions(ctx,line+pos+1);
	else if (!strcmp(buf,"match")) parse_rule(ctx,line+pos+1);
	else {
	    fprintf(stderr,"unrecognized line \"%s\"\n",line);
	    exit(1);
	}
    }
    return ctx;
}


static void destroy_pragmas(TCCEXT_PRAGMA *p)
{
    while (p) {
	TCCEXT_PRAGMA *next = p->next;

	free((char *) p->pragma);
	free(p);
	p = next;
    }
}


static void destroy_parameters(TCCEXT_PARAMETER *prm)
{
    while (prm) {
	TCCEXT_PARAMETER *next = prm->next;

	free((char *) prm->name);
	free(prm);
	prm = next;
    }
}


static void destroy_action(TCCEXT_ACTION *a)
{
    switch (a->type) {
	case tat_class:
	    while (a->u.classes) {
		TCCEXT_CLASS *next = a->u.classes->next;

		destroy_parameters(a->u.classes->parameters);
		free(a->u.classes);
		a->u.classes = next;
	    }
	    break;
	case tat_conform:
	    destroy_action(a->u.conform.yes_action);
	    destroy_action(a->u.conform.no_action);
	    break;
	case tat_count:
	    destroy_action(a->u.count.action);
	    break;
	default:
	    break;
    }
    free(a);
}


void tccext_destroy(TCCEXT_CONTEXT *ctx)
{
    destroy_pragmas(ctx->pragmas);
    while (ctx->buckets) {
	TCCEXT_BUCKET *next_bucket;

	destroy_pragmas(ctx->buckets->pragmas);
	next_bucket = ctx->buckets->next;
	free(ctx->buckets);
	ctx->buckets = next_bucket;
    }
    while (ctx->offset_groups) {
	TCCEXT_OFFSET *next_offset;

	next_offset = ctx->offset_groups->next;
	free(ctx->offset_groups);
	ctx->offset_groups = next_offset;
    }
    while (ctx->actions) {
	TCCEXT_ACTIONS *next_action;

	destroy_action(ctx->actions->action);
	next_action = ctx->actions->next;
	free(ctx->actions);
	ctx->actions = next_action;
    }
    while (ctx->rules) {
	TCCEXT_RULE *next_rule;

	while (ctx->rules->matches) {
	    TCCEXT_MATCH *next_match;

	    next_match = ctx->rules->matches->next;
	    free(ctx->rules->matches->data);
	    free(ctx->rules->matches);
	    ctx->rules->matches = next_match;
	}
	next_rule = ctx->rules->next;
	free(ctx->rules);
	ctx->rules = next_rule;
    }
    while (ctx->blocks) {
	TCCEXT_BLOCK *next_block;

	while (ctx->blocks->qdiscs) {
	    TCCEXT_QDISC *qdisc = ctx->blocks->qdiscs;
	    TCCEXT_QDISC *next_qdisc;

	    destroy_parameters(qdisc->parameters);
	    destroy_pragmas(qdisc->pragmas);
	    next_qdisc = qdisc->next;
	    free(qdisc);
	    ctx->blocks->qdiscs = next_qdisc;
	}
	destroy_pragmas(ctx->blocks->pragmas);
	next_block = ctx->blocks->next;
	free(ctx->blocks);
	ctx->blocks = next_block;
    }
}
