/*
 * ext_all.c - Dump entire configuration at external interface
 *
 * Written 2001 by Werner Almesberger 
 * Copyright 2001 Network Robots
 */

/*
 * This is experimental code and will change. Probably a lot.
 *
 * Known problems: @@@
 * - should allow simple configurations using other qdiscs than dsmark
 * - does not provide information on class hierarchy
 * - does not handle "pragma"
 * - prints flag parameters with numeric argument (is this a feature ?)
 * - does not support qdiscs shared by multiples classes
 * - gets confused if dumping more than one qdisc tree (ingress/egress) per
 *   session
 * - does not take into account that only part of the dsmark class may be used
 *   for marking
 * - does not take into account that no dsmark class may exist for marking
 *   (in fact, should look at parameters to decide whether a change is
 *   actually requested)
 * - does not report inner qdiscs changing skb->tc_index
 */


#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "config.h"
#include "error.h"
#include "param.h"
#include "tree.h"
#include "filter.h"
#include "qdisc.h"
#include "ext_all.h"


int dump_all; /* dump all qdiscs and filters (experimental) */


static const QDISC *_root; /* @@@@ HACK */


/* -------------------- skb->tc_index to class mapping --------------------- */


static void print_class_n(FILE *file,const QDISC *qdisc,uint16_t class,
  int *first)
{
    fprintf(file,"%c%u:%u",*first ? ' ' : ',',(unsigned) qdisc->number,
      (unsigned) class);
    *first = 0;
}


static void print_class(FILE *file,const CLASS *class,int *first)
{
    print_class_n(file,class->parent.qdisc,class->number,first);
}


static void gred_path(FILE *file,const QDISC *qdisc,uint16_t index,int first)
{
    const CLASS *class;

    param_get(qdisc->params);
    for (class = qdisc->classes; class; class = class->next)
	if (class->number == (index & 0xf)) {
	    print_class(file,class,&first);
	    return;
	}
    print_class_n(file,qdisc,prm_default_index.v,&first);
}


static void no_filters_at_classes(const CLASS *class)
{
    while (class) {
	if (class->filters)
	    yyerror("dump_all currently does not support filters at classes");
	no_filters_at_classes(class->child);
	class = class->next;
    }
}


static void find_path(FILE *file,const QDISC *qdisc,uint16_t index,int first);


static void filter_path(FILE *file,const QDISC *qdisc,uint16_t index,int first)
{
    const ELEMENT *element;
    int shift;

    if (!qdisc->filters) {
	if (qdisc->dsc->flags & QDISC_HAS_CLASSES)
	    yyerrorf("don't know how to select classes of \"%s\"",
	      qdisc->dsc->name);
	return;
    }
    if (qdisc->filters->next)
	yyerrorf("dump_all currently supports only one filter per qdisc");
    if (qdisc->filters->dsc != &tcindex_dsc)
	yyerrorf("dump_all supports only \"tcindex\", not \"%s\"",
	  qdisc->filters->dsc->name);
    param_get(qdisc->filters->params);
    shift = prm_shift.v;
    for (element = qdisc->filters->elements; element; element = element->next) {
	param_get(element->params);
	if (((index >> shift) & (prm_mask.present ? prm_mask.v : 0xffff)) ==
	  element->number) {
	    print_class(file,element->parent.class,&first);
	    if (element->parent.class->qdisc)
		find_path(file,element->parent.class->qdisc,index,first);
	    return;
	}
    }
    yyerror("dump_all found no matching choice in tcindex");
}


static void find_path(FILE *file,const QDISC *qdisc,uint16_t index,int first)
{
    no_filters_at_classes(qdisc->classes);
    if (qdisc->dsc == &gred_dsc) gred_path(file,qdisc,index,first);
    else filter_path(file,qdisc,index,first);
}


int dump_all_decision(FILE *file,uint16_t class)
{
    const CLASS *c;

    if (!dump_all) return 0;
    for (c = _root->classes; c; c = c->next)
	if (c->qdisc) {
	    int one = 1;

	    fprintf(file," class");
	    print_class_n(file,_root,class,&one);
	    find_path(file,c->qdisc,class,0);
	    return 1;
	}
    return 0;
}


/* ------------------- Dumps qdisc and class definitions ------------------- */


static void print_pragma(FILE *file,PARAM *params)
{
    const DATA_LIST *pragma;

    if (!prm_present(params,&prm_pragma)) return;
    fprintf(file," pragma");
    for (pragma = prm_data(params,&prm_pragma).u.list; pragma;
      pragma = pragma->next)
	fprintf(file," %s",pragma->ref->u.string);
}


static void print_class_qdisc(FILE *file,const PARAM_DSC *param,void *arg)
{
    const CLASS *class = arg;

    if (param != &prm_pragma)
	fprintf(file," qdisc %d",(int) class->qdisc->number);
}


static void dump_class(FILE *file,const CLASS *class)
{
    const QDISC *qdisc = class->parent.qdisc;
    static const PARAM_DSC *hide_pragma[] = {
	&prm_pragma,
	NULL
    };
    const PARAM_DSC **special = hide_pragma;
    void (*special_fn)(FILE *file,const PARAM_DSC *param,void *arg) = NULL;

    fprintf(file,"class %u =",(unsigned) class->number);
    /* @@@ <HACKS>*/
    if (qdisc->dsc == &gred_dsc) {
	static const PARAM_DSC *hide_default[] = {
	    &prm_default,
	    &prm_pragma,
	    NULL
	};

	special = hide_default;
	special_fn = NULL;
    }
    if (class->qdisc) {
	static PARAM_DSC pseudo_prm_qdisc = {
	    .id = "qdisc",
	};
	static const PARAM_DSC *class_qdisc[] = {
	    &pseudo_prm_qdisc,
	    &prm_pragma,
	    NULL
	};

	special = class_qdisc;
	special_fn = print_class_qdisc;
    }
    /* @@@ </HACKS>*/
    param_print(file,qdisc->dsc->qdisc_param,qdisc->dsc->class_param,
      qdisc->dsc->class_param,special,special_fn,(void *) class);
    print_pragma(file,class->params);
    fprintf(file,"\n");
}


static void dump_classes(FILE *file,const CLASS *class)
{
    while (class) {
	param_load(class->params,class->parent.qdisc->dsc->qdisc_param,
	  class->parent.qdisc->dsc->class_param);
	dump_class(file,class);
	param_push();
	dump_classes(file,class->child);
	param_pop();
	class = class->next;
    }
}


static void dump_this_qdisc(FILE *file,const QDISC *qdisc);


static void dump_qdiscs_at_classes(FILE *file,const CLASS *class)
{
    while (class) {
	if (class->qdisc) dump_this_qdisc(file,class->qdisc);
	dump_qdiscs_at_classes(file,class->child);
	class = class->next;
    }
}


static void gred_default(FILE *file,const PARAM_DSC *param,void *arg)
{
    const QDISC *qdisc = arg;
    const CLASS *class;

    if (param == &prm_pragma) return;
    for (class = qdisc->classes; class; class = class->next)
	if (prm_present(class->params,&prm_default)) {
	    fprintf(file," default %d",(int) class->number);
	    return;
	}
    abort();
}


static void prio_bands(FILE *file,const PARAM_DSC *param,void *arg)
{
    if (param != &prm_pragma) fprintf(file," bands 3");
}


static void dump_this_qdisc(FILE *file,const QDISC *qdisc)
{
    static const PARAM_DSC *hide_pragma[] = {
	&prm_pragma,
	NULL
    };
    const PARAM_DSC **special = hide_pragma;
    void (*special_fn)(FILE *file,const PARAM_DSC *param,void *arg) = NULL;

    fprintf(file,"qdisc %u = %s",(unsigned) qdisc->number,qdisc->dsc->name);
    param_load(qdisc->params,qdisc->dsc->qdisc_param,qdisc->dsc->qdisc_param);
    /* @@@ <HACKS>*/
    if (qdisc->dsc == &gred_dsc) {
	static const PARAM_DSC *insert_default[] = {
	    &prm_default,
	    NULL
	};

	special = insert_default;
	special_fn = gred_default;
    }
    if (qdisc->dsc == &prio_dsc) {
	static const PARAM_DSC *insert_bands[] = {
	    &prm_bands,
	    &prm_pragma,
	    NULL
	};

	if (!prm_present(qdisc->params,&prm_bands)) {
	    special = insert_bands;
	    special_fn = prio_bands;
	}
    }
    /* @@@ </HACKS>*/
    param_print(file,qdisc->dsc->qdisc_param,qdisc->dsc->class_param,
      qdisc->dsc->qdisc_param,special,special_fn,(void *) qdisc);
    print_pragma(file,qdisc->params);
    fprintf(file,"\n");
    param_push();
    dump_classes(file,qdisc->classes);
    param_pop();
    dump_qdiscs_at_classes(file,qdisc->classes);
}


/* ------------------------------------------------------------------------- */


void dump_all_qdisc(FILE *file,const QDISC *root)
{
    if (!root) yyerror("dump_all: no root qdisc");
    _root = root;
    if (root->filters) {
	if (root->filters->next) yyerror("dump_all: only one filter allowed");
	if (root->filters->dsc != &if_dsc)
	    yyerrorf("dump_all: filter must be \"if\", not \"%s\"",
	      root->filters->dsc->name);
    }
    if (root->dsc != &dsmark_dsc)
	yyerrorf("dump_all: root qdisc must be \"dsmark\", not \"%s\"",
	  root->dsc->name);
    dump_this_qdisc(file,root);
}
